# Copyright (c) 2016-2018  University of Houston

from __future__ import absolute_import

import os
import signal
import subprocess

import psutil

from .. import util
from ..util import KeyAdapter

class _AppMgr(object):
  PIDDIR = "/tmp/pids"

  def __init__ (self):
    try:
      os.makedirs(_AppMgr.PIDDIR)
    except OSError, e:
      if e.errno == 17:
        return
      raise e

  def writePID (self, app):
    with open("%s/%s-%d.pid" % (_AppMgr.PIDDIR, app.name, app._id), "w") as pidfile:
      pidfile.write("%d" % (app.pid))

  def delPID (self, app):
    os.remove("%s/%s-%d.pid" % (_AppMgr.PIDDIR, app.name, app._id))

ApplicationManager = _AppMgr()

def getOS ():
  name = None
  if os.path.exists("/etc/os-release"):
    with open("/etc/os-release", "r") as osfile:
      for line in osfile:
        if line.startswith("ID"):
          # Cut any whitespace, there should NOT be quotes
          name = line.split("=")[1].strip()
          return name.lower()
  elif os.path.exists("/etc/lsb-release"):
    with open("/etc/lsb-release", "r") as osfile:
      for line in osfile:
        if line.startswith("DISTRIB_ID"):
          # Cut any whitespace, there should NOT be quotes
          name = line.split("=")[1].strip()
          return name.lower()
  else:
    # For awful backwards compatibility
    return "ubuntu"

def getServiceManager ():
  cmdline = open("/proc/1/cmdline", "r").read()
  if cmdline.count("supervisord"):
    return "supervisor"
  elif cmdline.count("systemd"):
    return "systemd"
  elif cmdline.count("init"):
    p = subprocess.Popen("init --version", stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    (sout, serr) = p.communicate()
    lines = sout.split("\n")
    if lines:
      if lines[0].count("upstart"):
        return "upstart"
      elif lines[0].count("unrecognized"):
        real = os.path.realpath("/sbin/init")
        if real.count("systemd"):
          return "systemd"
        print real
    print sout

  return None



class LocalApplication(object):
  ID = 0

  def __init__ (self, name = None, logh = None):
    self._id = LocalApplication.ID
    LocalApplication.ID += 1
    self.name = name
    self._pid = None
    self.logh = KeyAdapter(name, logh)
    self._os = getOS()
    self._binpath_ubuntu = None
    self._binpath_alpine = None
    self._binpath_centos = None

  @property
  def binpath (self):
    path = getattr(self, "_binpath_%s" % (self._os))
    if not path:
      raise NotImplementedError()
    return path

  def start (self):
    pass

  def stop (self):
    pass

  def kill (self, sig = signal.SIGKILL):
    process = psutil.Process(self.pid)
    for proc in process.children(recursive=True):
      if self.logh: self.logh.debug("Killing %d (%s)" % (proc.pid, proc.name()))
      proc.send_signal(sig)
    if self.logh: self.logh.debug("Killing %d (%s)" % (self.pid, process.name()))
    process.send_signal(sig)
    if sig == signal.SIGKILL or sig == signal.SIGTERM:
      self._pid = None

  def call (self, cmd, block = True):
    if block:
      return util.call(cmd, self.logh, block = block)

    self._pid = util.call(cmd, self.logh, block = block)
    return self._pid

  @property
  def pid (self):
    return self._pid

  @pid.setter
  def pid (self, val):
    if isinstance(val, (str, unicode)):
      self._pid = int(val)
    else:
      self._pid = val

  @staticmethod
  def killall ():
    pass


class InstallableApplication(object):
  def __init__ (self, name, logh):
    self.name = name
    self.APT = "DEBIAN_FRONTEND=noninteractive apt-get install --no-install-recommends -y"
    self.APK = "apk add"
    self.logh = logh
    self._nonetwork = False

  @property
  def nonetwork (self):
    return self._nonetwork

  @nonetwork.setter
  def nonetwork (self, val):
    self._nonetwork = val

    # APT is the same command either way
    if self._nonetwork:
      self.APK = "apk add --no-network"
    else:
      self.APK = "apk add"

  def call (self, cmd):
    return util.call(cmd, logh = self.logh, block = True)

  def install (self, *args, **kwargs):
    os = getOS()

    inst = getattr(self, "install_%s" % (os))
    return inst(*args, **kwargs)

  def configure (self, *args, **kwargs):
    os = getOS()

    conf = getattr(self, "conf_%s" % (os))
    return conf(*args, **kwargs)

  def test (self, *args, **kwargs):
    os = getOS()

    tfunc = getattr(self, "test_%s" % (os))
    return tfunc(*args, **kwargs)

  def install_alpine (self, *args, **kwargs):
    raise NotImplementedError()

  def install_centos (self, *args, **kwargs):
    raise NotImplementedError()

  def install_ubuntu (self, *args, **kwargs):
    raise NotImplementedError()

  def conf_alpine (self, *args, **kwargs):
    raise NotImplementedError()

  def conf_centos (self, *args, **kwargs):
    raise NotImplementedError()

  def conf_ubuntu (self, *args, **kwargs):
    raise NotImplementedError()

  def test_alpine (self, *args, **kwargs):
    raise NotImplementedError()

  def test_centos (self, *args, **kwargs):
    raise NotImplementedError()

  def test_ubuntu (self, *args, **kwargs):
    raise NotImplementedError()


class LocalService(LocalApplication):
  def __init__ (self, name = None, logh = None):
    super(LocalService, self).__init__(name, logh)

  def start (self):
    svcmgr = getServiceManager()
    startf = getattr(self, "start_%s" % (svcmgr))
    return startf()

  def stop (self):
    svcmgr = getServiceManager()
    stopf = getattr(self, "stop_%s" % (svcmgr))
    return stopf()

  def start_supervisor (self):
    raise NotImplementedError()

  def start_systemd (self):
    raise NotImplementedError()

  def start_upstart (self):
    raise NotImplementedError()
