# Copyright (c) 2017-2018  University of Houston

from __future__ import absolute_import, print_function

import getpass
import json
import os.path
import sys
import time 

import paramiko
import paramiko.rsakey

from . import sshinfo
from ...util import KeyAdapter, PASSWD_ASK

class RemoteNodeProxy(object):
  def __init__ (self, logh):
    self.logh = logh
    self._ssh = None
    self._os = None
    self.run = None
    self.uhexp_path = None
    self._attrs = {}

  def checkPID (self, pid):
    return

  def downloadData (self, local_root, remote_root):
    return

  def run (self, cmd, timeout, pid, retry_count):
    return

  def getAttr (self, name, default = None):
    try:
      return self._attrs[name]
    except KeyError:
      return default


class StandardClient(RemoteNodeProxy):
  def __init__ (self, logh):
    super(StandardClient, self).__init__(logh)

  @property
  def ssh (self):
    return self._ssh

  @ssh.setter
  def ssh (self, val):
    self._ssh = val
    self.run = self._ssh.run

  @property
  def os (self):
    return self._os

  @os.setter
  def os (self, val):
    self._os = val
    self.logh.info("OS Detected: %s" % (val))

  def checkPID (self, pid):
    return not bool(int(self.run("test -d /proc/%d/; echo $?" % (pid))[0]))


class IGClient(RemoteNodeProxy):
  def __init__ (self, logh):
    super(IGClient, self).__init__(logh)
    self.host = None  # Friendly name
    self.context = None
    self.manifest_node = None
    self._uhexp_path = None

  @property
  def os (self):
    return self._os

  @os.setter
  def os (self, val):
    self._os = val
    self.logh.info("OS Detected: %s" % (val))

  @property
  def uhexp_path (self):
    return self._uhexp_path

  @uhexp_path.setter
  def uhexp_path (self, path):
    self._uhexp_path = path

  @property
  def ssh (self):
    return self._ssh

  @ssh.setter
  def ssh (self, val):
    self._ssh = val
    self.run = self._ssh.run

  def checkPID (self, pid):
    return not bool(int(self.run("test -d /proc/%d/; echo $?" % (pid))[0]))

  def downloadData (self, local_root, remote_root = "/exp-data"):
    return


class VTSClient(RemoteNodeProxy):
  def __init__ (self, logh):
    super(VTSClient, self).__init__(logh)
    self.host = None  # Friendly name
    self.context = None
    self.manifest_node = None
    self.site = None
    self.slicename = None
    self.project = None
    self._uhexp_path = None

  @property
  def os (self):
    return self._os

  @os.setter
  def os (self, val):
    self._os = val
    self.logh.info("OS Detected: %s" % (val))

  @property
  def uhexp_path (self):
    if not self._uhexp_path:
      for mount in self.manifest_node.mounts:
        if mount.type == "hg" and mount.name == "uhexp":
          self._uhexp_path = mount.path
    return self._uhexp_path

  @uhexp_path.setter
  def uhexp_path (self, path):
    self._uhexp_path = path

  @property
  def ssh (self):
    return self._ssh

  @ssh.setter
  def ssh (self, val):
    self._ssh = val
    self.run = self._ssh.run

  def checkPID (self, pid):
    return not bool(int(self.run("test -d /proc/%d/; echo $?" % (pid))[0]))

  def downloadData (self, local_root, remote_root = "/exp-data"):
    # TODO: We should probably run this in another process so the sleep for data copy doesn't block the next node
    self.context.project = self.project
    # TODO: This needs to use one of the dropbox mounts, instead of remote_root, maybe
    self.run("mkdir -p %s/_host_root" % (remote_root))
    self.run("cp *.log %(dl)s/_host_root/; cp *.out %(dl)s/_host_root/; cp *.json* %(dl)s/_host_root/" % ({"dl" : remote_root}))
    time.sleep(2)
    for mount in self.manifest_node.mounts:
      if mount.type == "dropbox":
        self.logh.info("Uploading data from %s:%s" % (self.host, mount.volid))
        self.site.dropboxUpload(self.context, self.slicename, [(self.host, mount.volid)])
    


def ClientFactory(log_handle, ssh_cfg_path = None, manifests = None, context_path = None,
                  ssh_pk_map = None, context = None, json_metadata_path = None):
  if not ssh_pk_map:
    ssh_pk_map = {}

  if ssh_cfg_path:
    config = paramiko.SSHConfig()
    config.parse(open(ssh_cfg_path))

    idl = []
    for host in config.get_hostnames():
      if host.count("*"):
        continue
      hdata = config.lookup(host)
      try:
        idl.extend(hdata["identityfile"])
      except KeyError:
        continue

    for id_path in idl:
      if id_path not in ssh_pk_map:
        with open(id_path, "r") as f:
          try:
            paramiko.rsakey.RSAKey.from_private_key(f)
            ssh_pk_map[id_path] = ""
          except:
            pw = getpass.getpass("[%s] Passphrase: " % (id_path))
            ssh_pk_map[id_path] = pw

  if manifests or context_path or context:
    return GENIClientFactory(log_handle, ssh_cfg_path, manifests, context_path, ssh_pk_map, context, json_metadata_path)
  elif json_metadata_path:
    return StandardFactory(log_handle, ssh_cfg_path, ssh_pk_map, json_metadata_path)

  
class UnknownClientError(Exception):
  def __init__ (self, name, factory):
    self.name = name
    self.factory = factory
  def __str__ (self):
    return "Requested client (%s) is not known to the Factory in use (%s)" % (self.name, self.factory)


class StandardFactory(object):
  def __init__ (self, log_handle, ssh_cfg_path, ssh_pk_map, json_metadata_path):
    self.logh = KeyAdapter("Standard Factory", log_handle)
    self.ssh_cfg_path = ssh_cfg_path
    self.ssh_pk_map = ssh_pk_map
    self.base_data = {}
    if json_metadata_path:
      self.base_data = json.loads(open(json_metadata_path, "rb").read())

  def build (self, name, hexdump = False):
    client = StandardClient(self.logh)
    for node in self.base_data["nodes"]:
      if node["name"] == name:
        for attr in node["attributes"]:
          client._attrs[attr["name"]] = attr["value"]

    ssh_client = sshinfo.CommonSSHInfo(self.logh, hexdump)
    ssh_client.host = name
    ssh_client.config_path = self.ssh_cfg_path
    ssh_client.pkey_passphrase = client.getAttr("ssh.password")

    client.host = name
    client.ssh = ssh_client
    client.uhexp_path = client.getAttr("uhexp.path")

    return client


class _GENIFactory(object):
  def __init__ (self, log_handle, metadata = None):
    self.ssh_cfg_path = None
    self.ssh_pk_map = {}
    self._manifests = []
    self.context = None
    self.metadata = metadata
    self.logh = KeyAdapter("GENI Factory", log_handle)

    self._client_map = {} # name : (manifest_node, project, slicename, am-name)

  @property
  def manifests (self):
    return self._manifests

  @manifests.setter
  def manifests (self, val):
    import geni.rspec.vtsmanifest
    import geni.rspec.pgmanifest

    self._client_map = {}
    for item in val:
      if isinstance(item, (geni.rspec.vtsmanifest.Manifest, geni.rspec.pgmanifest.Manifest)):
        m = item
      elif isinstance(item, (str, unicode)):
        m = geni.rspec.vtsmanifest.Manifest(item)
      else:
        self.logh.warning("Skipping manifest of unknown type: %s" % (type(item)))
        continue

      if isinstance(m, geni.rspec.vtsmanifest.Manifest):
        for node in m.containers:
          urn_project_slice = m.slicename.split(":")[-1]
          sparts = urn_project_slice.split("+")
          self._client_map[node.client_id] = (node, sparts[-3], sparts[-1], m.host)
      elif isinstance(m, geni.rspec.pgmanifest.Manifest):
        for node in m.nodes:
          self._client_map[node.client_id] = (node, None, None, None)
        

  def build (self, name, hexdump = False):
    import geni.aggregate.vts
    import geni.rspec.vtsmanifest
    import geni.rspec.pgmanifest

    try:
      (mnode, project, slicename, host) = self._client_map[name]
    except KeyError:
      raise UnknownClientError(name, "GENI")

    attr_data = {}
    if self.metadata:
      try:
        for node in self.metadata["nodes"]:
          if node["name"] == name:
            for attr in node["attributes"]:
              attr_data[attr["name"]] = attr["value"]
      except KeyError:
        pass
      
    if isinstance(mnode, geni.rspec.vtsmanifest.ManifestContainer):
      ssh_client = sshinfo.VTSSSHInfo(self.logh, hexdump)
      ssh_client.host = name
      ssh_client.config_path = self.ssh_cfg_path
      ssh_client.pkey_passphrase = self.ssh_pk_map[ssh_client.pkey_path]

      client = VTSClient(self.logh)
      client.host = name
      client.ssh = ssh_client
      client.manifest_node = mnode
      client.context = self.context
      client.site = geni.aggregate.vts.aggregateFromHost(host)
      client.slicename = slicename
      client.project = project
      client._attrs = attr_data

      return client

    elif isinstance(mnode, geni.rspec.pgmanifest.ManifestNode):
      ssh_client = sshinfo.CommonSSHInfo(self.logh, hexdump)
      ssh_client.host = name
      ssh_client.config_path = self.ssh_cfg_path
      ssh_client.pkey_passphrase = self.ssh_pk_map[ssh_client.pkey_path]

      client = IGClient(self.logh)
      client.host = name
      client.ssh = ssh_client
      client.manifest_node = mnode
      client.context = self.context
      client._attrs = attr_data
      client.uhexp_path = client.getAttr("uhexp.path")

      return client



def GENIClientFactory(log_handle, ssh_cfg_path = None, manifests = None,
                      context_path = None, ssh_pk_map = None, context = None, json_metadata_path = None):
  try:
    from geni.aggregate.frameworks import KeyDecryptionError
    import geni.util
  except ImportError:
    print("You broke it!  Install geni-lib!")
    sys.exit(1)

  metadata = None
  if json_metadata_path and os.path.exists(json_metadata_path):
    metadata = json.loads(open(json_metadata_path, "rb").read())

  GF = _GENIFactory(log_handle, metadata)
  GF.ssh_cfg_path = ssh_cfg_path

  need_pw = False
  if context_path or not context:
    try:
      context = geni.util.loadContext(path = context_path, key_passphrase = "ffffffffff")
    except TypeError:
      context = geni.util.loadContext(path = context_path)
    except KeyDecryptionError:
      need_pw = True

    if need_pw:
      while True:
        pw = getpass.getpass("GENI Certificate Private Key Passphrase: ")
        tries += 1
        try:
          context = geni.util.loadContext(path = context_path, key_passphrase = pw)
          break
        except KeyDecryptionError:
          if tries < 3:
            continue
          break
        except IOError:
          break

  GF.context = context

  if manifests:
    GF.manifests = manifests

  if ssh_pk_map:
    GF.ssh_pk_map = ssh_pk_map

  return GF

