# Copyright (c) 2017-2018  University of Houston

import getpass
import os
import paramiko
from paramiko.ssh_exception import SSHException
import socket
import string
import time

from ...util import KeyAdapter

class ExistingHostnameInConfigFileError(Exception):
  def __str__ (self):
    return "Hostname Exists: A SSH config file and host have already been provided"


class ExistingUsernameInConfigFileError(Exception):
  def __str__ (self):
    return "Username Exists: A SSH config file and host have already been provided"


class ExistingPortInConfigFileError(Exception):
  def __str__ (self):
    return "Port Exists: A SSH config file and host have already been provided"


class ExistingPkeyPathInConfigFileError(Exception):
  def __str__ (self):
    return "Identity File Exists: A SSH config file and host have already been provided"


class NoSSHInfoForHostError(Exception):
  def __str__ (self):
    return "There is no ssh config information for host"


class NoPassphraseForKeyError(Exception):
  def __str__ (self):
    return "No passphrase set for key"

class CommandTimeoutError(Exception):
  def __init__ (self, timeout, cmd, host):
    self.cmd = cmd
    self.host = host
    self.timeout = timeout

  def __str__ (self):
    return "Timed out (%ds) executing (%s) on (%s)" % (self.timeout, self.cmd, self.host)


class SSHInfo(object):
  def __init__ (self, logh, hexdump = False):
    self._config_path = None
    self._host = None
    self._hexdump = hexdump
    self._os = None

    self._hostname = None
    self._port = None
    self._username = None
    self._pkey_path = None

    self._pkey_passphrase = None

    self._client = None
    self._transport = None
    self._rawlogh = logh
    self.logh = KeyAdapter("SSH", logh)

  @property
  def host (self):
    return self._host

  @host.setter
  def host (self, val):
    if self._hostname: raise ExistingHostnameInConfigFileError
    if self._username: raise ExistingUsernameInConfigFileError
    if self._port: raise ExistingPortInConfigFileError
    if self._pkey_path: raise ExistingPkeyPathInConfigFileError
    self._host = val
    self.logh = KeyAdapter("SSH (%s)" % (self._host), self._rawlogh)

  @property
  def config_path (self):
    return self._config_path

  @config_path.setter
  def config_path (self, val):
    if self._hostname: raise ExistingHostnameInConfigFileError
    if self._username: raise ExistingUsernameInConfigFileError
    if self._port: raise ExistingPortInConfigFileError
    if self._pkey_path: raise ExistingPkeyPathInConfigFileError
    self._config_path = val

  @property
  def client (self):
    if not self._client:
      self._client = paramiko.SSHClient()
      self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
      self._client.set_log_channel("ssh.%s" % (self.host))
    return self._client

  def connect (self):
    if self._transport and self._transport.is_authenticated():
      return
    self.logh.debug("(Re)Connecting")
    self.client.connect(hostname = self.hostname, port = self.port,
                        username = self.username,
                        password = self.pkey_passphrase,
                        key_filename = self.pkey_path,
                        banner_timeout = 10)
    self._transport = self.client.get_transport()
    if self._hexdump:
      self._transport.set_hexdump(True)
#    self._transport.set_keepalive(1)

  def exec_command (self, cmd, get_pty=False, timeout=5):
    return self.client.exec_command(cmd, get_pty=get_pty, timeout=timeout)
    

  def _getDataFromConfig (self):
    config = paramiko.SSHConfig()
    config.parse(open(self._config_path))
    sshconf = config.lookup(self._host)
    self.logh.info("ssh config file with host %s and path %s" % (self._host, self._config_path))
    self._hostname = sshconf["hostname"]
    self._port = int(sshconf["port"])
    self._username = sshconf["user"]
    try:
      self._pkey_path = sshconf["identityfile"][0]
    except KeyError:
      self._pkey_path = None

  @property
  def hostname (self):
    if self._hostname:
      return self._hostname
    elif self._config_path and self._host:
      self._getDataFromConfig()
      return self._hostname
    else:
      raise NoSSHInfoForHostError

  @hostname.setter
  def hostname (self, val):
    if self.config_path or self.host:
      self.logh.error("[sshinfo] Hostname setter: %s" % (ExistingHostnameInConfigFileError))
      raise ExistingHostnameInConfigFileError
    self._hostname = val

  @property
  def port (self):
    if self._port:
      return self._port
    elif self._config_path and self._host:
      self._getDataFromConfig()
      return self._port
    else:
      raise NoSSHInfoForHostError

  @port.setter
  def port (self, val):
    if self.config_path or self.host:
      self.logh.error("[sshinfo] Port setter: %s" % (ExistingPortInConfigFileError))
      raise ExistingPortInConfigFileError
    self._port = val

  @property
  def username (self):
    if self._username:
      return self._username
    elif self._config_path and self._host:
      self._getDataFromConfig()
      return self._username
    else:
      raise NoSSHInfoForHostError

  @username.setter
  def username (self, val):
    if self.config_path or self.host:
      self.logh.error("Username setter: %s" % (ExistingUsernameInConfigFileError))
      raise ExistingUsernameInConfigFileError
    self._username = val

  @property
  def pkey_path (self):
    if self._pkey_path:
      return self._pkey_path
    elif self._config_path and self._host:
      self._getDataFromConfig()
      return self._pkey_path
    else:
      raise NoSSHInfoForHostError

  @pkey_path.setter
  def pkey_path (self, val):
    if self.config_path or self.host:
      self.logh.error("IdentityFile setter: %s" % (ExistingPkeyPathInConfigFileError))
      raise ExistingPkeyPathInConfigFileError
    self._pkey_path = val

  # TODO: what if a private key does not have a passphrase?
  @property
  def pkey_passphrase (self):
    if self._pkey_passphrase:
      return self._pkey_passphrase

  @pkey_passphrase.setter
  def pkey_passphrase (self, val):
    self._pkey_passphrase = val


class CommonSSHInfo(SSHInfo):
  def run (self, cmd, timeout = 5, pid = True, retry_count = 1):
    while retry_count >= 0:
      try:
        self.connect()

        try:
          if cmd.count("nohup"):
            newcmd = "%s; sleep 0.5" % (cmd)
          else:
            newcmd = cmd

          self.logh.info("Executing (%s)" % (newcmd))

          (i,o,e) = self.exec_command(newcmd, get_pty=True, timeout=timeout)
          data = o.read()
          err = e.read()
          if err:
            self.logh.warning("Output found on stderr:")
            self.logh.warning(err)
          return data.split("\r\n") if data else None
        except socket.timeout as ex:
          self.client.close()
          raise CommandTimeoutError(timeout, newcmd, self.host)

      except socket.error as e:
        self.logh.error("%s: %s trying again" % (cmd, e.message))
        retry_count -= 1
        time.sleep(0.2)
        if retry_count < 0:
          raise


class VTSSSHInfo(SSHInfo):
  def run (self, cmd, timeout = 5, pid = True, retry_count = 1):
    # there is no pid here but to have a 
    while retry_count >= 0:
      try:
        self.connect()

        try:
          if cmd.count("nohup"):
            newcmd = "%s; sleep 0.5" % (cmd)
#            newcmd = cmd
          else:
            newcmd = cmd

          self.logh.info("Executing (%s)" % (newcmd))

          (i,o,e) = self.exec_command(newcmd, get_pty=True, timeout=timeout)
          data = o.read()
          err = e.read()
          if err:
            self.logh.warning("Output found on stderr:")
            self.logh.warning(err)
          return data.split("\r\n") if data else None
        except socket.timeout as ex:
          self.client.close()
          raise CommandTimeoutError(timeout, newcmd, self.host)

      except socket.error as e:
        self.logh.error("%s: %s trying again" % (cmd, e.message))
        retry_count -= 1
        time.sleep(0.2)
        if retry_count < 0:
          raise
